@@ -5,14 +5,14 @@ import socket
 import struct
 import threading
 try:
-    from urllib import splitport
+    from urllib import splitport  # type: ignore
 except ImportError:
-    from urllib.parse import splitport
+    from urllib.parse import splitport  # type: ignore
 
 import zlib
+from io import BytesIO
 import six
 from six import binary_type, text_type
-import json
 
 from bmemcached.compat import long
 from bmemcached.exceptions import AuthenticationNotSupported, InvalidCredentials, MemcachedException
@@ -25,6 +25,26 @@ logger = logging.getLogger(__name__)
 class Protocol(threading.local):
     """
     This class is used by Client class to communicate with server.
+
+    Reference https://github.com/memcached/memcached/wiki/BinaryProtocolRevamped ::
+
+        Header structure
+        Byte/     0       |       1       |       2       |       3       |
+           /              |               |               |               |
+          |0 1 2 3 4 5 6 7|0 1 2 3 4 5 6 7|0 1 2 3 4 5 6 7|0 1 2 3 4 5 6 7|
+          +---------------+---------------+---------------+---------------+
+         0| Magic         | Opcode        | Key length                    |
+          +---------------+---------------+---------------+---------------+
+         4| Extras length | Data type     | vbucket id                    |
+          +---------------+---------------+---------------+---------------+
+         8| Total body length                                             |
+          +---------------+---------------+---------------+---------------+
+        12| Opaque                                                        |
+          +---------------+---------------+---------------+---------------+
+        16| CAS                                                           |
+          |                                                               |
+          +---------------+---------------+---------------+---------------+
+          Total 24 bytes
     """
     HEADER_STRUCT = '!BBHBBHLLQ'
     HEADER_SIZE = 24
@@ -79,7 +99,7 @@ class Protocol(threading.local):
     COMPRESSION_THRESHOLD = 128
 
     def __init__(self, server, username=None, password=None, compression=None, socket_timeout=None,
-                 dumps=None, loads=None):
+                 pickle_protocol=None, pickler=None, unpickler=None):
         super(Protocol, self).__init__()
         self.server = server
         self._username = username
@@ -89,8 +109,9 @@ class Protocol(threading.local):
         self.connection = None
         self.authenticated = False
         self.socket_timeout = socket_timeout
-        self.dumps = dumps
-        self.loads = loads
+        self.pickle_protocol = pickle_protocol
+        self.pickler = pickler
+        self.unpickler = unpickler
 
         self.reconnects_deferred_until = None
 
@@ -247,7 +268,7 @@ class Protocol(threading.local):
         if not self._username or not self._password:
             return False
 
-        logger.info('Authenticating as %s', self._username)
+        logger.debug('Authenticating as %s', self._username)
         self._send(struct.pack(self.HEADER_STRUCT,
                                self.MAGIC['request'],
                                self.COMMANDS['auth_negotiation']['command'],
@@ -268,7 +289,7 @@ class Protocol(threading.local):
 
         if b'PLAIN' not in methods:
             raise AuthenticationNotSupported('This module only supports '
-                                             'PLAIN auth for now.')
+                                             'PLAIN auth for now.', status)
 
         method = b'PLAIN'
         auth = '\x00%s\x00%s' % (self._username, self._password)
@@ -287,10 +308,10 @@ class Protocol(threading.local):
             return False
 
         if status == self.STATUS['auth_error']:
-            raise InvalidCredentials("Incorrect username or password")
+            raise InvalidCredentials("Incorrect username or password", status)
 
         if status != self.STATUS['success']:
-            raise MemcachedException('Code: %d Message: %s' % (status, extra_content))
+            raise MemcachedException('Code: %d Message: %s' % (status, extra_content), status)
 
         logger.debug('Auth OK. Code: %d Message: %s', status, extra_content)
 
@@ -323,10 +344,10 @@ class Protocol(threading.local):
             value = str(value)
         else:
             flags |= self.FLAGS['object']
-            dumps = self.dumps
-            if dumps is None:
-                dumps = self.json_dumps
-            value = dumps(value)
+            buf = BytesIO()
+            pickler = self.pickler(buf, self.pickle_protocol)
+            pickler.dump(value)
+            value = buf.getvalue()
 
         if compress_level != 0 and len(value) > self.COMPRESSION_THRESHOLD:
             if compress_level is not None and compress_level > 0:
@@ -366,29 +387,26 @@ class Protocol(threading.local):
         elif flags & FLAGS['long']:
             return long(value)
         elif flags & FLAGS['object']:
-            loads = self.loads
-            if loads is None:
-                loads = self.json_loads
-            return loads(value)
+            buf = BytesIO(value)
+            unpickler = self.unpickler(buf)
+            return unpickler.load()
 
         if six.PY3:
             return value.decode('utf8')
 
         # In Python 2, mimic the behavior of the json library: return a str
         # unless the value contains unicode characters.
+        # in Python 2, if value is a binary (e.g struct.pack("<Q") then decode will fail
         try:
             value.decode('ascii')
         except UnicodeDecodeError:
-            return value.decode('utf8')
+            try:
+                return value.decode('utf8')
+            except UnicodeDecodeError:
+                return value
         else:
             return value
 
-    def json_dumps(self, value):
-        return json.dumps(value).encode('utf8')
-
-    def json_loads(self, value):
-        return json.loads(value.decode('utf8'))
-
     def get(self, key):
         """
         Get a key and its CAS value from server.  If the value isn't cached, return
@@ -421,16 +439,48 @@ class Protocol(threading.local):
             if status == self.STATUS['server_disconnected']:
                 return None, None
 
-            raise MemcachedException('Code: %d Message: %s' % (status, extra_content))
+            raise MemcachedException('Code: %d Message: %s' % (status, extra_content), status)
 
         flags, value = struct.unpack('!L%ds' % (bodylen - 4, ), extra_content)
 
         return self.deserialize(value, flags), cas
 
+    def noop(self):
+        """
+        Send a NOOP command
+
+        :return: Returns the status.
+        :rtype: int
+        """
+        logger.debug('Sending NOOP')
+        data = struct.pack(self.HEADER_STRUCT +
+                           self.COMMANDS['noop']['struct'],
+                           self.MAGIC['request'],
+                           self.COMMANDS['noop']['command'],
+                           0, 0, 0, 0, 0, 0, 0)
+        self._send(data)
+
+        (magic, opcode, keylen, extlen, datatype, status, bodylen, opaque,
+         cas, extra_content) = self._get_response()
+
+        logger.debug('Value Length: %d. Body length: %d. Data type: %d',
+                     extlen, bodylen, datatype)
+
+        if status != self.STATUS['success']:
+            logger.debug('NOOP failed (status is %d). Message: %s' % (status, extra_content))
+
+        return int(status)
+
     def get_multi(self, keys):
         """
         Get multiple keys from server.
 
+        Since keys are converted to b'' when six.PY3 the keys need to be decoded back
+        into string . e.g key='test' is read as b'test' and then decoded back to 'test'
+        This encode/decode does not work when key is already a six.binary_type hence
+        this function remembers which keys were originally sent as str so that
+        it only decoded those keys back to string which were sent as string
+
         :param keys: A list of keys to from server.
         :type keys: list
         :return: A dict with all requested keys.
@@ -438,7 +488,8 @@ class Protocol(threading.local):
         """
         # pipeline N-1 getkq requests, followed by a regular getk to uncork the
         # server
-        keys, last = keys[:-1], keys[-1]
+        o_keys = keys
+        keys, last = keys[:-1], str_to_bytes(keys[-1])
         if six.PY2:
             msg = ''
         else:
@@ -454,7 +505,7 @@ class Protocol(threading.local):
                            self.COMMANDS['getk']['struct'] % (len(last)),
                            self.MAGIC['request'],
                            self.COMMANDS['getk']['command'],
-                           len(last), 0, 0, 0, len(last), 0, 0, last.encode())
+                           len(last), 0, 0, 0, len(last), 0, 0, last)
 
         self._send(msg)
 
@@ -468,11 +519,23 @@ class Protocol(threading.local):
                 flags, key, value = struct.unpack('!L%ds%ds' %
                                                   (keylen, bodylen - keylen - 4),
                                                   extra_content)
-                d[key.decode()] = self.deserialize(value, flags), cas
+                if six.PY2:
+                    d[key] = self.deserialize(value, flags), cas
+                else:
+                    try:
+                        decoded_key = key.decode()
+                    except UnicodeDecodeError:
+                        d[key] = self.deserialize(value, flags), cas
+                    else:
+                        if decoded_key in o_keys:
+                            d[decoded_key] = self.deserialize(value, flags), cas
+                        else:
+                            d[key] = self.deserialize(value, flags), cas
+
             elif status == self.STATUS['server_disconnected']:
                 break
             elif status != self.STATUS['key_not_found']:
-                raise MemcachedException('Code: %d Message: %s' % (status, extra_content))
+                raise MemcachedException('Code: %d Message: %s' % (status, extra_content), status)
 
         return d
 
@@ -519,7 +582,7 @@ class Protocol(threading.local):
                 return False
             elif status == self.STATUS['server_disconnected']:
                 return False
-            raise MemcachedException('Code: %d Message: %s' % (status, extra_content))
+            raise MemcachedException('Code: %d Message: %s' % (status, extra_content), status)
 
         return True
 
@@ -707,7 +770,7 @@ class Protocol(threading.local):
          cas, extra_content) = self._get_response()
 
         if status not in (self.STATUS['success'], self.STATUS['server_disconnected']):
-            raise MemcachedException('Code: %d Message: %s' % (status, extra_content))
+            raise MemcachedException('Code: %d Message: %s' % (status, extra_content), status)
         if status == self.STATUS['server_disconnected']:
             return 0
 
@@ -772,7 +835,7 @@ class Protocol(threading.local):
         if status == self.STATUS['server_disconnected']:
             return False
         if status != self.STATUS['success'] and status not in (self.STATUS['key_not_found'], self.STATUS['key_exists']):
-            raise MemcachedException('Code: %d message: %s' % (status, extra_content))
+            raise MemcachedException('Code: %d message: %s' % (status, extra_content), status)
 
         logger.debug('Key deleted %s', key)
         return status != self.STATUS['key_exists']
@@ -840,7 +903,7 @@ class Protocol(threading.local):
          cas, extra_content) = self._get_response()
 
         if status not in (self.STATUS['success'], self.STATUS['server_disconnected']):
-            raise MemcachedException('Code: %d message: %s' % (status, extra_content))
+            raise MemcachedException('Code: %d message: %s' % (status, extra_content), status)
 
         logger.debug('Memcached flushed')
         return True
